/*****************************************************************************
 *                                                                           *
 *      ***        ***                                                       *
 *      *****    *****                                                       *
 *      ******  ******                                                       *
 *      *** ****** ***                                                       *
 *      ***   **   ***                                                       *
 *      ***        ***                RFRAIN		  MAIN FILE              *
 *      ***        ***              RESPONSIVE ENVIRONMENTS GROUP            *
 *      ***        ***                      MIT MEDIA LAB                    *
 *      ***        ***                                                       *
 *      ***        ***                                                       *
 *                                                                           *
 *                                                                           *
 *****************************************************************************
 * This program implements a wireless network on the rf daughtercard         *
 * with a serial interface. It requires the CC1010 HAL (hardware             *
 * abstraction library, from chipcon) and the RFRAIN CUL (Chipcon utility    *
 * library, revision 2 by Mat Laibowitz based on chipcon's cul) to compile.  *
 *                                                                           *
 * Serial Protocol:                                                          *
 *  header byte 	-->	FFh                                                  *
 *  packet id		--> ID Byte
 *  opcode byte   	--> 01h  turn on rfdc		                             *
 *  							returns 00h if okay, FEh if error 			 *
 *						02h  turn off rfdc                                   *
 *  							returns 00h if okay, FEh if error 			 * 
 *						03h  set node options followed by 4 bytes            *
 *								Byte 1: node address MSB					 *
 *								Byte 2: node address LSB					 *
 *								Byte 3: Tx Ack Timeout						 *
 *								Byte 4: Tx Retry Count						 *
 *  							returns 00h if okay, FEh if error 			 *
 *           			04h  check status									 *
 *								returns status byte							 *
 *								d0: rx buffer contains new packet			 *
 *								d1: rx buffer full							 *
 *								d2: tx buffer full							 *
 *						05h  clear tx buffer								 *
 *  							returns 00h if okay, FEh if error 			 *
 *						06h  clear rx buffer								 *
 *  							returns 00h if okay, FEh if error 			 *
 *						07h  get packet from rx buffer						 *
 *								returns packet								 *
 *						08h  place packet in tx buffer followed by packet	 *
 *								returns 00h if acked or sent without ack req *
 *								returns FDh if tx buffer is full			 *
 *								returns FEh if transmission failed			 *
 *	packet		   -->	Destination address MSB								 *
 *						Destination address LSB								 *
 *						Data Length											 *
 *						Flags												 *
 *							d0: Acknowledge Request							 *
 *							d1: Ack flag (used internally, send as 0)  		 *
 *							d3: Encrypt packet (not yet implemented)		 *
 *							d4: Toggle bit (used internally, send as 0)		 *
 *							d5-d7: Unused									 *
 *						Data Bytes											 *						
 *	returns		  -->	FFh Header Byte										 *
 *						Packet ID											 *
 *						Opcode												 *
 *						Return code or packet								 *
 *																			 *
 * LEDs:                                                                     *
 *   White:  Successful transmission                                         *
 *   Blue:   Successful reception											 *
 *   Green:    Packet too long / received invalid ack                        *
 *****************************************************************************
 * Author:              Mat Laibowitz                                        *
 *****************************************************************************
 *																			 *
 * Revision history:                                                         *
 * 1.0  2003/07/31      First Public Release                                 *
 *                                                                           *
 ****************************************************************************/


#include <cc1010dc.h>
#include <cul.h>
#include <stdlib.h> // srand(), rand()
#include <string.h>

// Constant Definitions
#define	BUFFER_SIZE			128
#define UART_BUFFER_SIZE	64
#define MAX_PACKET_SIZE		32
#define LONG_RX_TIMEOUT		150
#define RSSI_THRESHOLD		0xBA
#define DEFAULT_ADDR_H		0x00
#define	DEFAULT_ADDR_L		0x06


// Global variables
SPP_SETTINGS xdata sppSettings;
SPP_RX_INFO xdata RXI;
SPP_TX_INFO xdata TXI;
byte xdata pRXData[MAX_PACKET_SIZE];
byte xdata pTXData[MAX_PACKET_SIZE];
byte xdata rx_buffer[BUFFER_SIZE];
byte xdata tx_buffer[BUFFER_SIZE];
byte xdata uart_buffer[UART_BUFFER_SIZE];
unsigned int xdata rx_tail, rx_head;
unsigned int xdata tx_tail, tx_head;
byte xdata uart_tail, uart_head, utx_active;
byte xdata in_packet[MAX_PACKET_SIZE];
byte xdata op_packet[MAX_PACKET_SIZE];
byte in_bytes,byte_index;
byte xdata tx_outgoing, current_tx_id;
byte xdata settingsRamArray[5];
byte code settingsFlashArray[5];
byte xdata ramBuffer[128];
byte rfdc_status;
byte header_rx;
byte op_state;
byte myRetryCount, txAttempts;
byte xdata setting_counter,setting_counter2,tx_counter,bad_bytes;
byte rf_on;
byte my_pkid;
byte last_packet[3];

// Function prototypes
unsigned char check_tx_buffer (byte length);
unsigned char check_rx_buffer (byte length);
void send_uart_byte (byte utx_byte);
void randomWait (void);
void initRSSI(void);
char rfdcReadRSSI(void); 

/*****************************************************************************
	MAIN PROGRAM
*****************************************************************************/
void main() {
	
#ifdef FREQ868

// X-tal frequency: 14.745600 MHz
// RF frequency A: 868.277200 MHz	Rx
// RF frequency B: 868.277200 MHz	Tx
// RX Mode: Low side LO
// Frequency separation: 64 kHz
// Data rate: 19.2 kBaud
// Data Format: NRZ
// RF output power: 4 dBm
// IF/RSSI: RSSI Enabled

RF_RXTXPAIR_SETTINGS code RF_SETTINGS = {
    0xA3, 0x2F, 0x15,    // Modem 0, 1 and 2
    0x75, 0xA0, 0x00,    // Freq A
    0x58, 0x32, 0x8D,    // Freq B
    0x01, 0xAB,          // FSEP 1 and 0
    0x40,                // PLL_RX
    0x30,                // PLL_TX
    0x6C,                // CURRENT_RX
    0xF3,                // CURRENT_TX
    0x32,                // FREND
    0xFF,                // PA_POW
    0x00,                // MATCH
    0x00,                // PRESCALER
    };

#endif

#ifdef FREQ915

// X-tal frequency: 14.745600 MHz
// RF frequency A: 915.027455 MHz	Rx
// RF frequency B: 915.027455 MHz	Tx
// RX Mode: Low side LO
// Frequency separation: 64 kHz
// Data rate: 19.2 kBaud
// Data Format: NRZ
// RF output power: 4 dBm
// IF/RSSI: RSSI Enabled

RF_RXTXPAIR_SETTINGS code RF_SETTINGS = {
    0xA3, 0x2F, 0x15,    // Modem 0, 1 and 2
    0xAA, 0x80, 0x00,    // Freq A
    0x5C, 0xF4, 0x02,    // Freq B
    0x01, 0xAB,          // FSEP 1 and 0
    0x58,                // PLL_RX
    0x30,                // PLL_TX
    0x6C,                // CURRENT_RX
    0xF3,                // CURRENT_TX
    0x32,                // FREND
    0xFF,                // PA_POW
    0x00,                // MATCH
    0x00,                // PRESCALER
    };


#endif

#ifdef FREQ433


// X-tal frequency: 14.745600 MHz
// RF frequency A: 433.302000 MHz	Rx
// RF frequency B: 433.302000 MHz	Tx
// RX Mode: Low side LO
// Frequency separation: 64 kHz
// Data rate: 19.2 kBaud
// Data Format: NRZ
// RF output power: 10 dBm
// IF/RSSI: RSSI Enabled

RF_RXTXPAIR_SETTINGS code RF_SETTINGS = {
    0x4B, 0x2F, 0x0E,    // Modem 0, 1 and 2
    0x58, 0x00, 0x00,    // Freq A
    0x41, 0xFC, 0x9C,    // Freq B
    0x02, 0x80,          // FSEP 1 and 0
    0x60,                // PLL_RX
    0x48,                // PLL_TX
    0x44,                // CURRENT_RX
    0x81,                // CURRENT_TX
    0x0A,                // FREND
    0xFF,                // PA_POW
    0xC0,                // MATCH
    0x00,                // PRESCALER
    };

#endif

	RF_RXTXPAIR_CALDATA xdata RF_CALDATA;

	int i;
	unsigned char tmp_RSSI, tmp_pid;
	byte q,p;
	unsigned int tx_tmp_head;


    // Set optimum settings for speed and low power consumption
    MEM_NO_WAIT_STATES();
    FLASH_SET_POWER_MODE(FLASH_STANDBY_BETWEEN_READS);
    
    // Initialize peripherals
    WDT_ENABLE(FALSE);
	WDT_DIVIDE_CLOCK_BY_16384();
    RLED_OE(TRUE);
    GLED_OE(TRUE);
    BLED_OE(TRUE);
    WLED_OE(TRUE);
    RLED = LED_OFF;
    GLED = LED_OFF;
	BLED = LED_OFF;
    WLED = LED_OFF;

	initRSSI();

    // Setup UART0 with interrupt
    UART0_SETUP(57600, CC1010DC_CLKFREQ, UART_NO_PARITY | UART_RX_TX | UART_ISR);
    INT_PRIORITY(INUM_UART0, INT_HIGH);
    INT_ENABLE(INUM_UART0, INT_ON);

	// Initialize variables
	tx_outgoing = 0;
	rfdc_status = 0;
	rx_tail = 0;
	rx_head = 0;
	tx_tail = 0;
	tx_head = 0;
	uart_tail = 0;
	uart_head = 0;
	header_rx = 0;
	utx_active = 0;
	op_state = 0;
	txAttempts = 0;
	rf_on = 0;
	tmp_RSSI = 0;
	my_pkid = 1;
	last_packet[0] = 0;
	last_packet[1] = 0;
	last_packet[2] = 0;


    // Seed the random generator:
    halRandomNumberGen(&q, 1);
    halRandomNumberGen(&p, 1);
    srand((q << 8) + p);
 
    memcpy(settingsRamArray,settingsFlashArray,5);
	if (settingsRamArray[0] != 0x6D) {
		settingsRamArray[0] = 0x6D;
		settingsRamArray[1] = DEFAULT_ADDR_H;
		settingsRamArray[2] = DEFAULT_ADDR_L;
		settingsRamArray[3] = 0x10;
		settingsRamArray[4] = 0x03;
		halCopy2Flash( settingsFlashArray, settingsRamArray, 5, ramBuffer, CC1010DC_CLKFREQ ); 
	}
   	sppSettings.myAddressH = settingsRamArray[1];
   	sppSettings.myAddressL = settingsRamArray[2];
   	sppSettings.txAckTimeout = settingsRamArray[3];
   	sppSettings.txAttempts = 1;
   	myRetryCount = settingsRamArray[4];
	sppSettings.rxTimeout = LONG_RX_TIMEOUT;
   	sppSettings.txPreambleByteCount = 7;

    // RF setup
    sppSetupRF(&RF_SETTINGS, &RF_CALDATA, TRUE);
    sppStartTimer(CC1010DC_CLKFREQ);
    SPP_INIT_TIMEOUTS();

    // Turn on interrupts
    INT_GLOBAL_ENABLE(INT_ON);

	rf_on = 1;

    RXI.pDataBuffer = pRXData;
	RXI.maxDataLen = MAX_PACKET_SIZE;			

	sppReceive(&RXI);
	setting_counter=0;
	tx_counter=0;	
	bad_bytes = 0;

    while (TRUE) {
		if (op_state & 0x01) {
			// SET RF SETTINGS
			op_state &= ~(0x01);
			halCopy2Flash( settingsFlashArray, settingsRamArray, 5, ramBuffer, CC1010DC_CLKFREQ ); 
		} else if (op_state & 0x02) {
			op_state &= ~(0x02);
			sppReset();
			sppReceive(&RXI);
			// RETURN CODE
			send_uart_byte(0xFF);
			send_uart_byte(op_packet[1]);
			send_uart_byte(0x01);
			send_uart_byte(0x00);
			tx_tail = tx_head = 0;
			rfdc_status &= ~(0x04);
			tx_outgoing = 0;
			rf_on = 1;
		} else if (op_state & 0x04) {
			op_state &= ~(0x04);
			sppReset();
			// RETURN CODE
			send_uart_byte(0xFF);
			send_uart_byte(op_packet[1]);
			send_uart_byte(0x02);
			send_uart_byte(0x00);
			tx_tail = tx_head = 0;
			rfdc_status &= ~(0x04);
			tx_outgoing = 0;
			rf_on = 0;
		} else if (op_state & 0x08) {
			op_state &= ~(0x08);
			tx_tail = tx_head = 0;
			rfdc_status &= ~(0x04);
			tx_outgoing = 0;
			send_uart_byte(0xFF);
			send_uart_byte(op_packet[1]);
			send_uart_byte(0x05);
			send_uart_byte(0x00);
		} else if (op_state & 0x10) {
			op_state &= ~(0x10);
			rx_tail = rx_head = 0;
			rfdc_status &= ~(0x01);
			rfdc_status &= ~(0x02);
			send_uart_byte(0xFF);
			send_uart_byte(op_packet[1]);
			send_uart_byte(0x06);
			send_uart_byte(0x00);
		}
	
	if (rf_on) {
    	if (RXI.status == SPP_RX_FINISHED) {
			// PACKET RECEIVED
			if ((RXI.sourceH != last_packet[0]) || (RXI.sourceL != last_packet[1]) || (RXI.pkid != last_packet[2])) {
			WLED = LED_ON;
			last_packet[0] = RXI.sourceH;
			last_packet[1] = RXI.sourceL;
			last_packet[2] = RXI.pkid;
			// CHECK IF THERE IS ROOM IN THE BUFFER
			if (check_rx_buffer(RXI.dataLen + 4) == 0) {
				// PUT PACKET IN BUFFER
				rx_buffer[rx_tail++] = RXI.sourceH;
				if (rx_tail == BUFFER_SIZE) rx_tail = 0;
				rx_buffer[rx_tail++] = RXI.sourceL;
				if (rx_tail == BUFFER_SIZE) rx_tail = 0;
				rx_buffer[rx_tail++] = RXI.dataLen;
				if (rx_tail == BUFFER_SIZE) rx_tail = 0;
				rx_buffer[rx_tail++] = RXI.flags;
				if (rx_tail == BUFFER_SIZE) rx_tail = 0;
				for (i=0;i<RXI.dataLen;i++) {
					rx_buffer[rx_tail++] = pRXData[i];
					if (rx_tail == BUFFER_SIZE) rx_tail = 0;
				}
				rfdc_status |= 0x01;
			} else {
				// RX BUFFER FULL
				rfdc_status |= 0x02;
			}
			}
			// RESTART RECEIVING
			sppReceive(&RXI);
		} else if (RXI.status == SPP_RX_TIMEOUT) {
			// RESTART RECEIVING
			sppReceive(&RXI);
		} else if (RXI.status == SPP_RX_TOO_LONG) {
			// RESTART RECEIVING
			sppReceive(&RXI);
		} else if (RXI.status == SPP_RX_RESET) {
			// RESTART RECEIVING
			sppReceive(&RXI);
		} else if (RXI.status == SPP_RX_WAITING) {
			// MEDIUM FREE AT A PROTOCOL LEVEL
			if (tx_outgoing == 1) {
				// CHECK RSSI
				tmp_RSSI = rfdcReadRSSI();
				if (tmp_RSSI < RSSI_THRESHOLD){
					halWait(rand(), CC1010DC_CLKFREQ);	
					if (txAttempts == 0){
						tx_tmp_head = tx_head;
						if (++tx_tmp_head == BUFFER_SIZE) tx_tmp_head = 0;
						tmp_pid = tx_buffer[tx_tmp_head]; 
						if (++tx_tmp_head == BUFFER_SIZE) tx_tmp_head = 0;
						if (++tx_tmp_head == BUFFER_SIZE) tx_tmp_head = 0;
            			TXI.destinationH = tx_buffer[tx_tmp_head];
						if (++tx_tmp_head == BUFFER_SIZE) tx_tmp_head = 0;
            			TXI.destinationL = tx_buffer[tx_tmp_head];
						if (++tx_tmp_head == BUFFER_SIZE) tx_tmp_head = 0;
            			TXI.dataLen = tx_buffer[tx_tmp_head];
						if (++tx_tmp_head == BUFFER_SIZE) tx_tmp_head = 0;
            			TXI.flags = tx_buffer[tx_tmp_head];
						if (++tx_tmp_head == BUFFER_SIZE) tx_tmp_head = 0;							
						for (i=0;i<TXI.dataLen;i++) {
							pTXData[i] = tx_buffer[tx_tmp_head];
							if (++tx_tmp_head == BUFFER_SIZE) tx_tmp_head = 0;
	            		}
						txAttempts = myRetryCount;
						TXI.pkid = ++my_pkid;
					}			
					TXI.pDataBuffer = pTXData;
					tmp_RSSI = rfdcReadRSSI();
					if ((RXI.status == SPP_RX_WAITING) && (tmp_RSSI < RSSI_THRESHOLD)) {
							sppReset();
							if (sppSend(&TXI) == SPP_TX_STARTED) {
								BLED = LED_ON;
					        	do {  } while (SPP_STATUS() != SPP_IDLE_MODE);
					            if (TXI.status == SPP_TX_FINISHED) {
									GLED = LED_ON;
									rfdc_status &= ~(0x04);
									tx_head = tx_tmp_head;
									if (tx_head == tx_tail) tx_outgoing = 0;
			            			txAttempts = 0;
									send_uart_byte(0xFF);
									send_uart_byte(tmp_pid);
									send_uart_byte(0x08);
									send_uart_byte(0x00);
                				} else {
									if (--txAttempts == 0) {
										RLED = LED_ON;
										rfdc_status &= ~(0x04);
										tx_head = tx_tmp_head;
										if (tx_head == tx_tail) tx_outgoing = 0;
			            				send_uart_byte(0xFF);
 										send_uart_byte(tmp_pid);
										send_uart_byte(0x08);
										send_uart_byte(0xFE);
									}
								}

							} 

					// RESTART RECEIVING
					sppReceive(&RXI);
 					} 
				} 
			}
		} 	

	}
		    RLED = LED_OFF;
    		GLED = LED_OFF;
			BLED = LED_OFF;
    		WLED = LED_OFF;
	}
} // main


// ISR (interrupt service routine) for uart0
void isr_uart0() interrupt INUM_UART0
{
	int i;
    byte received_byte;
	byte tmp_length;

    if (INT_GETFLAG(INUM_UART0_RX) == INT_SET) {
        received_byte = UART0_RECEIVE(); //read receive buffer
        INT_SETFLAG(INUM_UART0_RX, INT_CLR); // clear int
		if (header_rx == 0) {
			if (received_byte == 0xFF) {
				header_rx = 1;
				in_packet[0] = received_byte;
			} else {
				bad_bytes++;
			}
		} else if (header_rx == 1) {
			// PACKET ID RECEIVED
			header_rx = 2;
			in_packet[1] = received_byte;
		} else if (header_rx == 2) {
			// OPCODE RECEIVED
			in_packet[2] = received_byte;
			if(op_state > 0) {
				header_rx = 2;
			}
			if (received_byte == 0x01) {
					// TURN ON RX
					memcpy(&op_packet[1],&in_packet[1],1);
					op_state |= 0x02;
					header_rx = 0;
			} else if (received_byte == 0x02) {
					// TURN OFF RF
					memcpy(&op_packet[1],&in_packet[1],1);
					op_state |= 0x04;
					header_rx = 0;
			} else if (received_byte == 0x03) {
					// SETTINGS
					in_bytes = 4;
					header_rx = 3;
					byte_index = 3;		
			} else if (received_byte == 0x04) {
					// STATUS READ
					send_uart_byte(0xFF);
					send_uart_byte(in_packet[1]);
					send_uart_byte(0x04);
					send_uart_byte(rfdc_status);
					header_rx = 0;
			} else if (received_byte == 0x05) {
					// CLEAR TX BUFFER
					memcpy(&op_packet[1],&in_packet[1],1);
					op_state |= 0x08;
					header_rx = 0;
			} else if (received_byte == 0x06) {
					// CLEAR RX BUFFER
					memcpy(&op_packet[1],&in_packet[1],1);
					op_state |= 0x10;
					header_rx = 0;
			} else if (received_byte == 0x07) {
					// GET RX PACKET
					if (rfdc_status & 0x01){				
						send_uart_byte(0xFF);
						send_uart_byte(in_packet[1]);
						send_uart_byte(0x07);
						send_uart_byte(rx_buffer[rx_head]);
						if((++rx_head)==BUFFER_SIZE) rx_head = 0;
						send_uart_byte(rx_buffer[rx_head]);
						if((++rx_head)==BUFFER_SIZE) rx_head = 0;
						tmp_length = rx_buffer[rx_head];
						send_uart_byte(tmp_length);
						if((++rx_head)==BUFFER_SIZE) rx_head = 0;
						send_uart_byte(rx_buffer[rx_head]);
						if((++rx_head)==BUFFER_SIZE) rx_head = 0;
						for(i=0;i<tmp_length;i++) {
							send_uart_byte(rx_buffer[rx_head]);
							if((++rx_head)==BUFFER_SIZE) rx_head = 0;
						}
						rfdc_status &= ~(0x02);
						if (rx_head == rx_tail) rfdc_status &= ~(0x01);
					}
					header_rx = 0;
			} else if (received_byte == 0x08) {
					// PLACE PACKET IN TX BUFFER
					in_bytes = 4;
					header_rx = 3;
					byte_index = 3;
			} else {
					header_rx = 0;
			}
		} else if (header_rx == 3) {
			in_packet[byte_index] = received_byte;
			byte_index++;
			if ((--in_bytes) == 0) {
				if (in_packet[2] == 0x03) {
					setting_counter++;
					settingsRamArray[0] = 0x6D;
					settingsRamArray[1] = in_packet[3];
					settingsRamArray[2] = in_packet[4];
					settingsRamArray[3] = in_packet[5];
					settingsRamArray[4] = in_packet[6];
		   			sppSettings.myAddressH = settingsRamArray[1];
 					sppSettings.myAddressL = settingsRamArray[2];
					sppSettings.txAckTimeout = settingsRamArray[3];
					myRetryCount = settingsRamArray[4];
					send_uart_byte(0xFF);
					send_uart_byte(in_packet[1]);
					send_uart_byte(0x03);
					send_uart_byte(0x00);
					op_state |= 0x01;
					header_rx = 0;
				} else if (in_packet[2] == 0x08){
					if (in_packet[5]) {
						in_bytes = in_packet[5];
						header_rx = 4;
					} else {
						tx_counter++;	
						if (check_tx_buffer(in_packet[5]+7) == 0){
						// PUT in_packet in TX BUFFER
							for (i=0;i<(7+in_packet[5]);i++) {
								tx_buffer[tx_tail++] = in_packet[i];
								if (tx_tail == BUFFER_SIZE) tx_tail = 0;
							}
							tx_outgoing = 1;
						} else {
							// RETURN BUFFER FULL MESSAGE
							rfdc_status |= 0x04;
							send_uart_byte(0xFF);
							send_uart_byte(in_packet[1]);
							send_uart_byte(0x08);
							send_uart_byte(0xFD);
						}
						header_rx = 0;
					}
				}
			}
		} else if (header_rx == 4) {
			in_packet[byte_index] = received_byte;
			byte_index++;
			if ((--in_bytes) == 0) {
				tx_counter++;	
				if (check_tx_buffer(in_packet[5]+7) == 0){
					// PUT in_packet in TX BUFFER
					for (i=0;i<(7+in_packet[5]);i++) {
						tx_buffer[tx_tail++] = in_packet[i];
						if (tx_tail == BUFFER_SIZE) tx_tail = 0;
					}
					tx_outgoing = 1;
				} else {
					// RETURN BUFFER FULL MESSAGE
					rfdc_status |= 0x04;
					send_uart_byte(0xFF);
					send_uart_byte(in_packet[1]);
					send_uart_byte(0x08);
					send_uart_byte(0xFD);
				}
				header_rx = 0;
			}
		}				

    } else if (INT_GETFLAG(INUM_UART0_TX) == INT_SET) {
        INT_SETFLAG(INUM_UART0_TX, INT_CLR); // clear int
		if (uart_tail != uart_head) {
    		//BUFFER IS NOT EMPTY -- SEND NEXT BYTE
    		SBUF0 = uart_buffer[uart_head++];
			utx_active = 1;
    		if (uart_head == UART_BUFFER_SIZE) uart_head = 0;
  		} else {
			utx_active = 0;
		}

    }  

} //end isr_uart0()
// Flash interrupt handler (do nothing)
// We need to handle the interrupt even though we do not do anything.
// If not, the program will not run correctly except under the debugger,
// which has its own Flash interrupt handler

void FlashIntrHandler(void) interrupt INUM_FLASH 
{    
    INT_SETFLAG(INUM_FLASH, INT_CLR);    
	return;
}

unsigned char check_rx_buffer (byte length) 
{
	int new_tail;
//	char tmp_ret;
	
//	tmp_ret = 0x00;
	new_tail = rx_tail + length;

	if (new_tail >= BUFFER_SIZE) {
		new_tail = new_tail - BUFFER_SIZE;
		if((rx_tail < rx_head) || (new_tail >= rx_head)) {
			//tmp_ret = 0xFE;
			return (254);
		} 
	}

	if (rx_tail < rx_head) {
		if (new_tail >= rx_head) {
			//tmp_ret = 0xFE;
			return (254);
		}
	}
	return (0);
}

unsigned char check_tx_buffer (byte length) 
{
	int new_tail;
//	char tmp_ret;

//	tmp_ret = 0x00;
	new_tail = tx_tail + length;

	if (new_tail >= BUFFER_SIZE) {
		new_tail = new_tail - BUFFER_SIZE;
		if((tx_tail < tx_head) || (new_tail >= tx_head)){
			//tmp_ret = 0xFE;			
			return (254);
		}
	}

	if (tx_tail < tx_head) {
		if (new_tail >= tx_head){
			//tmp_ret = 0xFE;
			return (254);
		}
	}
	return (0);
}

void send_uart_byte (byte utx_byte)
{
//	unsigned char buf_empty = 0;
// 	if( uart_tail == uart_head ) buf_empty = 1;
  	uart_buffer[uart_tail++] = utx_byte;
  	if (uart_tail == UART_BUFFER_SIZE) uart_tail = 0;
  	if (utx_active == 0) {
  		utx_active = 1;
   		SBUF0 = uart_buffer[uart_head++];
   		if (uart_head == UART_BUFFER_SIZE) uart_head = 0;
//    	INT_SETFLAG(INUM_UART0_TX, INT_SET); 	//MANUALLY TRIGGER INTERRUPT
//  	} else {
//		FREND |= 0x02;
	}
}



void initRSSI(void)
{
    FREND|=0x02;                // Activate RSSI output

    #if (CC1010DC_CLKFREQ == 14746)
    ADCON2=0x03;                // ADC divider i 1/64 (optimal for 14.7456 MHz
                                // CLK), Interrupt not enabled.
    #else
    ADCON2=0x02;                // ADC divider i 1/48 (optimal for 11.0592 MHz
                                // CLK), Interrupt not enabled.
    #endif

    ADCON=0x0A;                 // ADC power on, single-conversion mode, 
                                // input from AD2 pin
    halWait(200, CC1010DC_CLKFREQ);
}

char rfdcReadRSSI(void) 
{
	byte val;

    ADC_RUN(TRUE);
	
    while (ADCON&0x04);         // Wait until sample is available
    
    val=ADC_GET_SAMPLE_8BIT();  // Get 8 MSB of sample
    
    //ADC_RUN(FALSE);           // Power down ADC
    
    val = -50-((val*55)>>8);
    return val;
}    
